# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from __future__ import division, print_function, unicode_literals

import shlex

from ..imports import *


class soundcloud_backend(MediaHost):

    DOMAINS = ('soundcloud.com',)
    NAME = 'Soundcloud'

    def init(self):
        client_id = self.settings.get('SOUNDCLOUD_CLIENT_ID', '')
        if not client_id:
            return
        self.info_url = 'http://api.soundcloud.com/resolve.json' \
                      + '?client_id='+client_id \
                      + '&url=http://soundcloud.com/'

    def get_id(self, url):
        return url.path[1:]

    def get_info(self, track_id, raw):
        if not hasattr(self, 'info_url'):
            raise MediaHostException(self._('Error: you need to provide a client ID to use the %s backend') % self.NAME)
        r = iNS(type='audio')
        info = iNS(requests.get(self.info_url+track_id).json())
        if raw:
            return info
        if info.kind != 'track':
            raise MediaHostException(self._('%s is a %s, not a track' % (track_id,info.kind)))
        r.alternates = call(lambda url: [{'type':'video','url':url}], info.video_url or identity)
        make_author = safe(lambda u: {'name':u['username'], 'urlname':u['permalink']})
        r.authors = call(lambda a: [a], make_author(info.user or {}))
        r.artworks = call(lambda url: [{'url':url}], info.artwork_url)
        r.beats_per_minute = info.bpm or identity
        r.comment_count = info.comment_count
        r.description = info.description
        r.duration = info.duration / 1000
        r.favorite_count = info.favoritings_count
        r.genres = call(lambda a: [a], info.genre or identity)
        r.license = info.license
        r.published = call(parse_date, info.created_at)
        r.purchase_url = info.purchase_url or identity
        r.tags = shlex.split(info.tag_list)
        r.thumbnails = call(lambda url: [{'url': url}], info.waveform_url)
        r.title = info.title
        r.subtype = info.track_type or identity
        r.view_count = info.playback_count
        r.downloads = list(filter(None, [info.download_url, info.stream_url]))
        return r

    def normalize(self, track_id):
        return 'http://soundcloud.com/'+track_id



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .six.moves import reduce


def getattrs(o, *attrs, **kwargs):
    """
    >>> getattrs((), '__iter__', '__name__', 'strip')('_')
    'iter'
    >>> getattrs((), 'foo', 'bar', default=0)
    0
    """
    if 'default' in kwargs:
        default = kwargs['default']
        c = o
        for attr in attrs:
            try:
                c = getattr(c, attr)
            except AttributeError:
                return default
        return c
    else:
        return reduce(getattr, attrs, o)



# This program is placed into the public domain.

__all__ = ('get_version')

from os.path import dirname, isdir, join
import re
from subprocess import CalledProcessError, check_output

version_re = re.compile('^Version: (.+)$', re.M)

def get_version():
    d = dirname(__file__)

    if isdir(join(d, '.git')):
        # Get the version using "git describe".
        cmd = 'git describe --tags --match [0-9]*'.split()
        try:
            version = check_output(cmd).decode().strip()
        except CalledProcessError:
            print('Unable to get version number from git tags')
            exit(1)

        # PEP 386 compatibility
        if '-' in version:
            version = '.post'.join(version.split('-')[:2])

    else:
        # Extract the version from the PKG-INFO file.
        with open(join(d, 'PKG-INFO')) as f:
            version = version_re.search(f.read()).group(1)

    return version


if __name__ == '__main__':
    print(get_version())



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


"""
Utilities for writing code that runs on Python 2 and 3.

Original code from http://packages.python.org/six
"""

import operator
import sys
import types


# True if we are running on Python 3.
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,
    integer_types = int,
    class_types = type,
    ustr = text_type = str
    binary_type = bytes

    MAXSIZE = sys.maxsize
else:
    string_types = basestring,
    integer_types = (int, long)
    class_types = (type, types.ClassType)
    ustr = text_type = unicode
    binary_type = str

    if sys.platform == 'java':
        # Jython always uses 32 bits.
        MAXSIZE = int((1 << 31) - 1)
    else:
        # It's possible to have sizeof(long) != sizeof(Py_ssize_t).
        class X(object):
            def __len__(self):
                return 1 << 31
        try:
            len(X())
        except OverflowError:
            # 32-bit
            MAXSIZE = int((1 << 31) - 1)
        else:
            # 64-bit
            MAXSIZE = int((1 << 63) - 1)
            del X


def _add_doc(func, doc):
    """Add documentation to a function."""
    func.__doc__ = doc


def _import_module(name):
    """Import module, returning the module after the last dot."""
    __import__(name)
    return sys.modules[name]


class _LazyDescr(object):

    def __init__(self, name):
        self.name = name

    def __get__(self, obj, tp):
        result = self._resolve()
        setattr(obj, self.name, result)
        # This is a bit ugly, but it avoids running this again.
        delattr(tp, self.name)
        return result


class MovedModule(_LazyDescr):

    def __init__(self, name, old, new=None):
        super(MovedModule, self).__init__(name)
        if PY3:
            if new is None:
                new = name
            self.mod = new
        else:
            self.mod = old

    def _resolve(self):
        return _import_module(self.mod)


class MovedAttribute(_LazyDescr):

    def __init__(self, name, old_mod, new_mod, old_attr=None, new_attr=None):
        super(MovedAttribute, self).__init__(name)
        if PY3:
            if new_mod is None:
                new_mod = name
            self.mod = new_mod
            if new_attr is None:
                if old_attr is None:
                    new_attr = name
                else:
                    new_attr = old_attr
            self.attr = new_attr
        else:
            self.mod = old_mod
            if old_attr is None:
                old_attr = name
            self.attr = old_attr

    def _resolve(self):
        module = _import_module(self.mod)
        return getattr(module, self.attr)



class _MovedItems(types.ModuleType):
    """Lazy loading of moved objects"""


_moved_attributes = [
    MovedAttribute('cStringIO', 'cStringIO', 'io', 'StringIO'),
    MovedAttribute('filter', 'itertools', 'builtins', 'ifilter', 'filter'),
    MovedAttribute('input', '__builtin__', 'builtins', 'raw_input', 'input'),
    MovedAttribute('map', 'itertools', 'builtins', 'imap', 'map'),
    MovedAttribute('reload_module', '__builtin__', 'imp', 'reload'),
    MovedAttribute('reduce', '__builtin__', 'functools'),
    MovedAttribute('StringIO', 'StringIO', 'io'),
    MovedAttribute('xrange', '__builtin__', 'builtins', 'xrange', 'range'),
    MovedAttribute('zip', 'itertools', 'builtins', 'izip', 'zip'),

    MovedModule('builtins', '__builtin__'),
    MovedModule('configparser', 'ConfigParser'),
    MovedModule('copyreg', 'copy_reg'),
    MovedModule('http_cookiejar', 'cookielib', 'http.cookiejar'),
    MovedModule('http_cookies', 'Cookie', 'http.cookies'),
    MovedModule('html_entities', 'htmlentitydefs', 'html.entities'),
    MovedModule('html_parser', 'HTMLParser', 'html.parser'),
    MovedModule('http_client', 'httplib', 'http.client'),
    MovedModule('BaseHTTPServer', 'BaseHTTPServer', 'http.server'),
    MovedModule('CGIHTTPServer', 'CGIHTTPServer', 'http.server'),
    MovedModule('SimpleHTTPServer', 'SimpleHTTPServer', 'http.server'),
    MovedModule('cPickle', 'cPickle', 'pickle'),
    MovedModule('queue', 'Queue'),
    MovedModule('reprlib', 'repr'),
    MovedModule('socketserver', 'SocketServer'),
    MovedModule('tkinter', 'Tkinter'),
    MovedModule('tkinter_dialog', 'Dialog', 'tkinter.dialog'),
    MovedModule('tkinter_filedialog', 'FileDialog', 'tkinter.filedialog'),
    MovedModule('tkinter_scrolledtext', 'ScrolledText', 'tkinter.scrolledtext'),
    MovedModule('tkinter_simpledialog', 'SimpleDialog', 'tkinter.simpledialog'),
    MovedModule('tkinter_tix', 'Tix', 'tkinter.tix'),
    MovedModule('tkinter_constants', 'Tkconstants', 'tkinter.constants'),
    MovedModule('tkinter_dnd', 'Tkdnd', 'tkinter.dnd'),
    MovedModule('tkinter_colorchooser', 'tkColorChooser',
                'tkinter.colorchooser'),
    MovedModule('tkinter_commondialog', 'tkCommonDialog',
                'tkinter.commondialog'),
    MovedModule('tkinter_tkfiledialog', 'tkFileDialog', 'tkinter.filedialog'),
    MovedModule('tkinter_font', 'tkFont', 'tkinter.font'),
    MovedModule('tkinter_messagebox', 'tkMessageBox', 'tkinter.messagebox'),
    MovedModule('tkinter_tksimpledialog', 'tkSimpleDialog',
                'tkinter.simpledialog'),
    MovedModule('urllib_robotparser', 'robotparser', 'urllib.robotparser'),
    MovedModule('winreg', '_winreg'),
]
for attr in _moved_attributes:
    setattr(_MovedItems, attr.name, attr)
del attr

moves = sys.modules['miss.six.moves'] = _MovedItems('moves')


def add_move(move):
    """Add an item to six.moves."""
    setattr(_MovedItems, move.name, move)


def remove_move(name):
    """Remove item from six.moves."""
    try:
        delattr(_MovedItems, name)
    except AttributeError:
        try:
            del moves.__dict__[name]
        except KeyError:
            raise AttributeError('no such move, %r' % (name,))


if PY3:
    _meth_func = '__func__'
    _meth_self = '__self__'

    _func_code = '__code__'
    _func_defaults = '__defaults__'

    _iterkeys = 'keys'
    _itervalues = 'values'
    _iteritems = 'items'
else:
    _meth_func = 'im_func'
    _meth_self = 'im_self'

    _func_code = 'func_code'
    _func_defaults = 'func_defaults'

    _iterkeys = 'iterkeys'
    _itervalues = 'itervalues'
    _iteritems = 'iteritems'


try:
    advance_iterator = next
except NameError:
    def advance_iterator(it):
        return it.next()
next = advance_iterator


if PY3:
    def get_unbound_function(unbound):
        return unbound

    Iterator = object

    def callable(obj):
        return any('__call__' in klass.__dict__ for klass in type(obj).__mro__)
else:
    def get_unbound_function(unbound):
        return unbound.im_func

    class Iterator(object):

        def next(self):
            return type(self).__next__(self)

    callable = callable
_add_doc(get_unbound_function,
         """Get the function out of a possibly unbound function""")


get_method_function = operator.attrgetter(_meth_func)
get_method_self = operator.attrgetter(_meth_self)
get_function_code = operator.attrgetter(_func_code)
get_function_defaults = operator.attrgetter(_func_defaults)


def iterkeys(d):
    """Return an iterator over the keys of a dictionary."""
    return iter(getattr(d, _iterkeys)())

def itervalues(d):
    """Return an iterator over the values of a dictionary."""
    return iter(getattr(d, _itervalues)())

def iteritems(d):
    """Return an iterator over the (key, value) pairs of a dictionary."""
    return iter(getattr(d, _iteritems)())


if PY3:
    def b(s):
        return s.encode('latin-1')
    def u(s):
        return s
    if sys.version_info[1] <= 1:
        def int2byte(i):
            return bytes((i,))
    else:
        # This is about 2x faster than the implementation above on 3.2+
        int2byte = operator.methodcaller('to_bytes', 1, 'big')
    import io
    StringIO = io.StringIO
    BytesIO = io.BytesIO
else:
    def b(s):
        return s
    def u(s):
        return unicode(s, 'unicode_escape')
    int2byte = chr
    import StringIO
    StringIO = BytesIO = StringIO.StringIO
_add_doc(b, """Byte literal""")
_add_doc(u, """Text literal""")


if PY3:
    import builtins
    exec_ = getattr(builtins, 'exec')


    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value


    print_ = getattr(builtins, 'print')
    del builtins

else:
    def exec_(code, globs=None, locs=None):
        """Execute code in a namespace."""
        if globs is None:
            frame = sys._getframe(1)
            globs = frame.f_globals
            if locs is None:
                locs = frame.f_locals
            del frame
        elif locs is None:
            locs = globs
        exec('exec code in globs, locs')


    exec_('''def reraise(tp, value, tb=None):
    raise tp, value, tb
''')


    def print_(*args, **kwargs):
        """The new-style print function."""
        fp = kwargs.pop('file', sys.stdout)
        if fp is None:
            return
        def write(data):
            if not isinstance(data, basestring):
                data = str(data)
            fp.write(data)
        want_unicode = False
        sep = kwargs.pop('sep', None)
        if sep is not None:
            if isinstance(sep, unicode):
                want_unicode = True
            elif not isinstance(sep, str):
                raise TypeError('sep must be None or a string')
        end = kwargs.pop('end', None)
        if end is not None:
            if isinstance(end, unicode):
                want_unicode = True
            elif not isinstance(end, str):
                raise TypeError('end must be None or a string')
        if kwargs:
            raise TypeError('invalid keyword arguments to print()')
        if not want_unicode:
            for arg in args:
                if isinstance(arg, unicode):
                    want_unicode = True
                    break
        if want_unicode:
            newline = unicode('\n')
            space = unicode(' ')
        else:
            newline = '\n'
            space = ' '
        if sep is None:
            sep = space
        if end is None:
            end = newline
        for i, arg in enumerate(args):
            if i:
                write(sep)
            write(arg)
        write(end)

_add_doc(reraise, """Reraise an exception.""")


def with_metaclass(meta, base=object):
    """Create a base class with a metaclass."""
    return meta('NewBase', (base,), {})



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from __future__ import print_function, unicode_literals

import argparse
import json
try:
    import jsonpickle
except ImportError:
    jsonpickle = None

from . import MediaHosts, MediaHostException
from .utils import urldecode


class FakeEncoder(json.JSONEncoder):
    def default(self, obj):
        try:
            return json.JSONEncoder.default(self, obj)
        except TypeError:
            return str(obj)


p = argparse.ArgumentParser()
p.add_argument('uri')
p.add_argument('settings', nargs='*', default=[], help='e.g. SOUNDCLOUD_CLIENT_ID=x')
p.add_argument('-b', '--backends', nargs='*', default=None)
p.add_argument('-r', '--raw', default=False, action='store_true')
args = p.parse_args()


settings = dict(s.split('=') for s in args.settings)
media_hosts = MediaHosts(backends=args.backends, settings=settings)
try:
    info = media_hosts.get_info_by_url(args.uri, args.raw)
except MediaHostException as e:
    print(e.args[0])
    exit(1)

if jsonpickle:
    print(json.dumps(json.loads(jsonpickle.encode(info.__dict__)), indent=4))
else:
    print(json.dumps(info.__dict__, indent=4, cls=FakeEncoder))



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from __future__ import division, print_function, unicode_literals

from ..imports import *


track_id_re = re.compile(r'/track/[^/]*/([0-9]+)')

class beatport_backend(MediaHost):

    DOMAINS = ('beatport.com',)
    NAME = 'Beatport'

    def get_id(self, url):
        track_id = track_id_re.search(url.path)
        if not track_id:
            raise MediaHostException(self._('Unrecognized url: %s') % url.geturl())
        return track_id.group(1)

    def get_info(self, track_id, raw):
        r = iNS(type='audio')
        info_url = 'http://api.beatport.com/catalog/3/beatport/track?id='
        info = iNS(requests.get(info_url+track_id).json())
        if raw:
            return info
        t = iNS(info.results.get('track', None))
        if not t:
            raise MediaHostException(self._('Failed to retrieve information for track %s' % (track_id)))
        make_author = safe(lambda a: {'name':a['name'], 'urlname':a['slug']})
        r.authors = list(filter(None, map(make_author, t.artists)))
        waveform = t.images.pop('waveform', identity)
        r.artworks = list(t.images.values())
        r.beats_per_minute = t.bpm or identity
        r.clef = t.key.pop('standard')
        r.clef.update(t.key)
        minutes, seconds = t.length.split(':')
        r.duration = int(minutes)*60 + int(seconds)
        r.genres = list(filter(None, (g.get('name', None) for g in t.genres)))
        r.published = call(parse_date, t.publishDate or identity)
        r.thumbnails = call(lambda a: [a], waveform)
        r.title = t.title
        return r

    def normalize(self, track_id):
        return 'http://www.beatport.com/track//'+track_id



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


def pstrip(s, prefix):
    """Prefix strip

    >>> pstrip('foo_oof', 'foo_')
    'oof'
    >>> pstrip('baroof', 'foo_')
    'baroof'
    """
    i = len(prefix)
    if s[:i] == prefix:
        return s[i:]
    return s

def sstrip(s, suffix):
    """Suffix strip

    >>> sstrip('foo.oof', '.oof')
    'foo'
    >>> sstrip('baroof', '.oof')
    'baroof'
    """
    i = - len(suffix)
    if s[i:] == suffix:
        return s[:i]
    return s



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .six.moves import reduce


inf = float('inf')
neg_inf = float('-inf')

def multiply(*xs):
    """
    >>> multiply(1, 2, 3)
    6
    """
    return reduce(lambda x, y: x*y, xs)



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .six.moves import reduce


def concat(l):
    """
    >>> concat([[0, 1], [2], [3, 4, 5]])
    [0, 1, 2, 3, 4, 5]
    """
    return reduce(list.__add__, l, [])

def grow(l, size, filler=None):
    """Grow the list $l to $size filling with $filler.
    >>> grow(list(range(5)), 4)
    [0, 1, 2, 3, 4]
    >>> grow(list(range(5)), 5)
    [0, 1, 2, 3, 4]
    >>> grow(list(range(5)), 6)
    [0, 1, 2, 3, 4, None]
    >>> grow(list(range(5)), 7, lambda i: i)
    [0, 1, 2, 3, 4, 5, 6]
    """
    ll = len(l)
    if ll < size:
        if callable(filler):
            return l + [filler(i) for i in range(ll, size)]
        else:
            return l + [filler]*(size-ll)
    else:
        return l



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from __future__ import division, print_function, unicode_literals

import traceback

from ..imports import *


def text(o, *keys):
    keys += ('$t',)
    return getitemsi(o, *keys)


video_id_re = re.compile(r'/(?:watch\?.*v=)?([a-zA-Z0-9_-]{11})')

class youtube_backend(MediaHost):

    DOMAINS = ('youtube.com', 'youtu.be')
    NAME = 'Youtube'

    def get_id(self, url):
        video_id = video_id_re.search(url.path + '?' + url.query)
        if not video_id:
            raise MediaHostException(self._('Unrecognized url: %s') % url.geturl())
        return video_id.group(1)

    def get_info(self, video_id, raw):
        r = iNS(type='video')
        errors = []

        # Try to get info from /get_video_info
        info_url = 'http://youtube.com/get_video_info?video_id=%s' % video_id
        try:
            info = urldecode(requests.get(info_url).text)
            if not info:
                errors.append(self._('Received empty response from %s') % info_url)
        except Exception:
            traceback.print_exc()
            info = {}
        if raw:
            pass
        elif info.get('status', 'ok') != 'ok':
            errors.append(info.get('reason', ustr(info)).replace('\\n', '\n'))
        elif info:
            info = iNS(info)
            r.duration = int(info.length_seconds)
            r.published = call(datetime.fromtimestamp, int(info.timestamp))
            r.rating = call(lambda a: dict(average=a), float(info.avg_rating))
            r.tags = info.keywords.split(',')
            r.thumbnails = call(lambda url: [{'url': url}], info.thumbnail_url)
            r.title = info.title
            r.view_count = int(info.view_count)
            # Decode download info
            ds = info.url_encoded_fmt_stream_map.split(',')
            ds = {d['itag']: d for d in map(urldecode, ds) if 'itag' in d}
            for fmt in info.fmt_list.split(','):
                itag, resolution, extra_data = fmt.split('/', 2)
                # TODO figure out what extra_data is
                d = ds[itag]
                d.update(resolution=resolution, extra_data=extra_data)
                d['yt_format'] = d.pop('itag')
            r.downloads = ds.values()

        # Try to get info from gdata API v2
        url = 'http://gdata.youtube.com/feeds/api/videos/%s?v=2&alt=json' % video_id
        try:
            entry = requests.get(url).json()['entry']
            if not entry:
                errors.append(self._('Received empty response from %s') % url)
        except Exception:
            traceback.print_exc()
            entry = {}
        if raw:
            return iNS(get_video_info=info, gdata=entry)
        elif entry:
            i = iNS(entry)
            media = iNS(i['media$group'])
            stats = iNS(i['yt$statistics'])
            make_author = safe(lambda a:
                {'name': author.name.text,
                 'urlname': author.uri.text.split('/')[-1]}
            )
            r.authors = list(filter(None, map(make_author, i.author))) or r.authors
            r.description = text(media, 'media$description')
            r.duration = getitemsi(media, 'yt$duration', 'seconds')
            r.favorite_count = stats['favoriteCount']
            r.published = call(parse_date, text(i, 'published'), r.published)
            r.rating = i['gd$rating']
            r.rating['num_raters'] = r.rating.pop('numRaters', identity)
            r.recorded = call(parse_date, text(i, 'yt$recorded'))
            r.thumbnails += media['media$thumbnail']
            r.title = text(i, 'title') or r.title
            r.updated = call(parse_date, text(i, 'updated'))
            r.view_count = stats['viewCount'] or r.view_count
            d = [dict(url=c.url, type=c.type, yt_format=c['yt$format'])
                 for c in map(iNS, media['media$content'])
                 if c.type.startswith('video')]
            r.downloads += d

        if not r.title:
            if errors:
                error = self._('Youtube said:\n"%s"') % max(errors, key=len)
            else:
                error = self._('unknown reason')
            raise MediaHostException(self._('Getting video info failed, %s') % error)

        return r

    def normalize(self, video_id):
        """
        >>> print(normalize(mirror, urlparse('http://youtube.com/watch?a=b&v=0aA1bB2cC9z&c=d')))
        http://youtube.com/watch?v=0aA1bB2cC9z
        >>> print(normalize(mirror, urlparse('http://youtu.be/0aA1bB2cC9z')))
        http://youtube.com/watch?v=0aA1bB2cC9z
        """
        return 'http://youtube.com/watch?v='+video_id



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from datetime import *


def age(born):
    today = date.today()
    try: # February 29 issue
        anniversary = born.replace(year=today.year)
    except ValueError:
        anniversary = born.replace(year=today.year, day=1, month=3)
    if anniversary > today:
        return today.year - born.year - 1
    else:
        return today.year - born.year



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .identity import identity


class NS(object):
    """Namespace

    >>> ns = NS({'a': 0}, {'b': 0}, b=1)
    >>> ns
    NS({'a': 0, 'b': 1})
    >>> assert NS(['a', 'b'], 0, b=1) == ns
    >>> assert NS(ns) == ns
    >>> assert NS(ns) is not ns
    """
    def __init__(self, *args, **kwargs):
        if args:
            if isinstance(args[0], NS):
                for ns in args:
                    self.__dict__.update(ns.__dict__)
            elif isinstance(args[0], dict):
                for d in args:
                    self.__dict__.update(d)
            else:
                self.__dict__.update(dict.fromkeys(*args))
        if kwargs:
            self.__dict__.update(kwargs)
    def __contains__(self, item):
        return item in self.__dict__
    def __eq__(self, other):
        return self.__dict__ == other
    def __getitem__(self, key):
        return self.__dict__[key]
    def __iter__(self):
        return iter(self.__dict__)
    def __len__(self):
        return len(self.__dict__)
    def __setitem__(self, key, value):
        self.__dict__[key] = value
    def __repr__(self):
        return self.__class__.__name__+'('+repr(self.__dict__)+')'
    def __str__(self):
        return (
            self.__class__.__name__+
            '('+
            ', '.join(str(k)+'='+str(v) for k, v in self.__dict__.items())+
            ')'
        )
    def _clear(self):
        self.__dict__.clear()
    def _get(self, key, default=None):
        return self.__dict__.get(key, default)
    def _pop(self, *args, **kwargs):
        return self.__dict__.pop(*args, **kwargs)
    def _update(self, *args, **kwargs):
        return self.__dict__.update(*args, **kwargs)
    def _setdefault(self, key, default):
        return self.__dict__.setdefault(key, default)


class iNS(NS):
    """Identity Namespace

    When accessing a nonexistent attribute, returns the Identity instead of
    raising an exception.
    """
    __getitem__ = __getattr__ = \
        lambda self, name: self.__dict__.get(name, identity)
    def __setattr__(self, key, value):
        if value != identity:
            self.__dict__[key] = value
    __setitem__ = __setattr__
    def _pop(self, key, default=identity):
        return self.__dict__.pop(key, default)



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .collections import UserDict
from .six import PY3


class lazy_attr(object):
    """A decorator for lazy attributes.

    >>> class A(object):
    ...     @lazy_attr
    ...     def a(self):
    ...         return ['a']
    >>> a = A()
    >>> assert a.a is a.a  # check that evaluation only occurs once
    >>> a.a
    ['a']
    """

    def __init__(self, func):
        self.__func = func
        self.__name__ = func.__name__
        self.__module__ = func.__module__
        self.__doc__ = func.__doc__
        self.__dict__.update(func.__dict__)

    def __get__(self, inst, owner):
        if inst is None:
            return self

        name = self.__name__
        if name.startswith('__') and not name.endswith('__'):
            name = '_%s%s' % (owner.__name__, name)

        value = self.__func(inst)
        setattr(inst, name, value)
        return value


class lazy(object):
    """A basic lambda wrapper.

    >>> a = lazy(lambda: 'foo')
    >>> a()
    'foo'
    >>> assert a() is a()
    """

    def __init__(self, func):
        self.func = func

    __call__ = lambda self: self.value

    @lazy_attr
    def value(self):
        v = self.func()
        self.func = None
        return v


class lazy_dict(UserDict):
    """A lazy dictionary.

    >>> d = lazy_dict(dict(a=lambda:'a'))
    >>> d.pop('a')
    'a'
    >>> d.lazy('b', lambda: 1)
    >>> d
    {'b': 1}
    """
    def __init__(self, lambdas_dict, *args, **kwargs):
        kwargs.update((k, lazy(v)) for k, v in lambdas_dict.items())
        UserDict.__init__(self, *args, **kwargs)
    def __getitem__(self, key):
        o = UserDict.__getitem__(self, key)
        if isinstance(o, lazy):
            o = self[key] = o.value
        return o
    def __repr__(self):
        return (
            '{'+
            ', '.join(repr(k)+': '+repr(v) for k, v in self.items())+
            '}'
        )
    def lazy(self, key, func):
        self[key] = lazy(func)



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


try:
    from urllib.parse import parse_qsl
except ImportError:
    from urlparse import parse_qsl

from miss.six import PY3


if PY3:
    def urldecode(s):
        if isinstance(s, bytes):
            s = s.decode()
        return {k: len(v) == 1 and v[0] or v
                for k, v in parse_qsl(s)}
else:
    def udecode(o, encoding='utf8'):
        if isinstance(o, list):
            return [udecode(s, encoding) for s in o]
        else:
            return o.decode(encoding)
    def urldecode(s):
        return {k: udecode(len(v) == 1 and v[0] or v)
                for k, v in parse_qsl(s)}



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from datetime import datetime
import re
try:
    from urllib.parse import *
except ImportError:
    from urlparse import *

try:
    from dateutil.parser import parse as parse_date
except ImportError:
    parse_date = lambda _: identity
import requests

from miss.identity import *
from miss.namespace import iNS
from miss.six import ustr

from . import MediaHost, MediaHostException, mirror
from .utils import *



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


def compose(f, g, unpack=False):
    """
    >>> compose(int, str)(0)
    0
    >>> compose(lambda *a: a, lambda *a: a, True)(0, 1)
    (0, 1)
    """
    if unpack:
        return lambda *args, **kwargs: f(*g(*args, **kwargs))
    else:
        return lambda *args, **kwargs: f(g(*args, **kwargs))



# -*- coding: utf-8 -*-

from os.path import join, dirname

from setuptools import setup, find_packages

from version import get_version

setup(
    name='media-hosts',
    version=get_version(),
    description='Gets media info from sites like Youtube and Soundcloud',
    author='Changaco',
    author_email='changaco ατ changaco δοτ net',
    url='https://github.com/Changaco/python-media-hosts',
    license='LGPLv3+',
    packages=find_packages(),
    long_description=open(join(dirname(__file__), 'README.rst')).read(),
    install_requires = [
        'miss>=0.3',
        'requests',
    ],
)



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from .lazy import lazy
from .six import PY3
from .six.moves import reduce


def _derive(cls):
    c = lazy(lambda: type('Identity', (Identity, cls), {}))
    return lambda self: c()(cls)

class Identity(object):
    """The identity element of many basic python operations.

    >>> i = Identity()
    >>> i
    Identity()
    >>> (i + 1) + (2 * i) + (3**i) + (i - 4j) + (i / 2)
    (6-4j)
    >>> i & 1 | 2
    2
    >>> i + []
    []
    >>> print(i.foo()['bar'])
    ?

    Be careful with id comparison ("is" operator),
    we store multiple instances to fake numeric types.
    >>> i is Identity()
    True
    >>> int(i) is int(i)
    True
    >>> i is int(i)
    False

    You can use value comparison instead, it will only return True
    when compared with other instances of Identity subclasses.
    >>> i == int(i)
    True
    >>> i == object()
    False
    """
    def __new__(cls, sup=object):
        if '_inst' in cls.__dict__:
            return cls._inst
        r = cls._inst = sup.__new__(cls)
        return r
    def __getattribute__(self, name):
        if name[0] == '_':
            return object.__getattribute__(self, name)
        return self
    __abs__ = __and__ = __call__ = __delattr__ = __delitem__ = \
    __div__ = __floordiv__ = __getattr__ = __getitem__ = \
    __invert__ = __lshift__ = __mod__ = __neg__ = __pos__= __pow__ = \
    __rand__ = __rshift__ = __setattr__ = __setitem__ = __truediv__ = \
        lambda self, *args, **kwargs: self
    __add__ = __mul__ = __or__ = __radd__ = __rdiv__ = __rfloordiv__ = \
    __rlshift__ = __rrshift__ = __rmod__ = __rmul__ = __ror__ = \
    __rpow__ = __rsub__ = __rtruediv__ = __rxor__ = __xor__ = \
        lambda self, other: other
    __bool__ = __nonzero__ = lambda self: False
    __eq__ = lambda self, other: issubclass(other.__class__, Identity)
    __iter__ = lambda self: iter(())
    __len__ = lambda self: 0
    __repr__ = lambda self: 'Identity()'
    __str__ = lambda self: '?'
    __sub__ = lambda self, other: - other
    __complex__ = _derive(complex)
    __float__ = _derive(float)
    __int__ = __round__ = _derive(int)
    if PY3:
        __bytes__ = lambda self: b'?'
    else:
        __long__ = _derive(long)
        __unicode__ = lambda self: unicode('?')

identity = Identity()


def call(func, obj, default=identity):
    """
    Return func(obj) if obj != identity else return default.
    """
    if obj != identity:
        return func(obj)
    return default


def call_pop(func, d, key, default=identity):
    """
    Return func(d.pop(key)) if key in d, else return default.
    """
    if key in d:
        return func(d.pop(key))
    return default


def getattri(o, attr):
    return getattr(o, attr, identity)


def getattrsi(o, *attrs):
    return reduce(lambda a, b: getattr(a, b, identity), attrs, o)


def getitemi(o, key):
    try:
        return o[key]
    except KeyError:
        return identity


def getitemsi(o, *keys):
    return reduce(lambda a, b: getitemi(a, b), keys, o)


def safe(f, default=identity):
    """
    Run function f ignoring all exceptions.
    """
    def g(*args, **kwargs):
        try:
            return f(*args, **kwargs)
        except:
            return default
    return g



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.

"""
This module is a version of collections consistent in both python 2 and 3.

It is only made of re-exports from builtin modules, with one exception. Unlike
the builtin UserDict, the one in this module derives from dict, which means that
existing code using `isintance(obj, dict)` won't break.
"""

from collections import *

from .six import PY3

if PY3:
    class UserDict(UserDict, dict): pass
else:
    import UserDict as _UserDict
    class UserDict(_UserDict.DictMixin, dict):
        __contains__ = dict.__contains__
        __iter__ = dict.__iter__
    from UserList import *
    from UserString import *



# This file is part of a program licensed under the terms of the GNU Lesser
# General Public License version 3 (or at your option any later version)
# as published by the Free Software Foundation.
#
# If you have not received a copy of the GNU Lesser General Public License
# along with this file, see <http://www.gnu.org/licenses/>.


from importlib import import_module
from os import listdir
from os.path import dirname
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from miss.str import pstrip


mirror = lambda a: a


def list_backends():
    return [a[:-3] for a in listdir(dirname(__file__)+'/backends')
                   if a[0] != '_' and a.endswith('.py')]


class MediaHosts:

    def __init__(self, backends=None, _=mirror, settings={}):
        self._ = _
        self.backends = []
        for backend in (backends or list_backends()):
            try:
                module = import_module('media_hosts.backends.'+backend)
                cls = getattr(module, backend+'_backend')
                instance = cls(_, settings)
                self.backends.append(instance)
            except Exception:
                print('Failed to import "%s" backend:' % backend)
                import traceback
                traceback.print_exc()
        self.domains = {domain: backend for backend in self.backends
                                        for domain in backend.DOMAINS}

    def get_backend(self, url_str):
        url = urlparse(url_str) # this doesn't validate the URL and doesn't throw exceptions
        try:
            return (url, self.domains[pstrip(url.hostname, 'www.')])
        except KeyError:
            raise MediaHostException(self._('Unrecognized domain name: %s') % url.hostname)

    def get_info_by_id(self, backend, media_id, raw=False):
        info = backend.get_info(media_id, raw)
        info.id = media_id
        info.service = backend.NAME
        return info

    def get_info_by_url(self, url_str, raw=False):
        url, backend = self.get_backend(url_str)
        return self.get_info_by_id(backend, backend.get_id(url), raw)

    def normalize(self, url_str):
        url, backend = self.get_backend(url_str)
        i = backend.get_id(url)
        return (backend, backend.normalize(i), i)


class MediaHost(object):

    def __init__(self, _, settings):
        self._ = _
        self.settings = settings
        self.init()

    def init(self):
        pass


class MediaHostException(Exception): pass
